94057c3d3e
PR Test (NPU) / check-changes (push) Has been cancelled
PR Test (NPU) / pr-gate (push) Has been cancelled
PR Test (NPU) / set-image-config (push) Has been cancelled
PR Test (NPU) / stage-b-test-1-npu-a2 (0) (push) Has been cancelled
PR Test (NPU) / stage-b-test-1-npu-a2 (1) (push) Has been cancelled
PR Test (NPU) / stage-b-test-2-npu-a2 (0) (push) Has been cancelled
PR Test (NPU) / stage-b-test-2-npu-a2 (1) (push) Has been cancelled
PR Test (NPU) / stage-b-test-4-npu-a3 (push) Has been cancelled
PR Test (NPU) / stage-b-test-16-npu-a3 (push) Has been cancelled
PR Test (NPU) / multimodal-gen-test-1-npu-a3 (push) Has been cancelled
PR Test (NPU) / multimodal-gen-test-2-npu-a3 (push) Has been cancelled
PR Test (Arm64) / pr-gate (push) Has been cancelled
PR Test (Arm64) / check-changes (push) Has been cancelled
PR Test (Arm64) / build-test (push) Has been cancelled
PR Test (sgl-router) / gate (push) Has been cancelled
PR Test (sgl-router) / tier-1 — lint (push) Has been cancelled
PR Test (sgl-router) / tier-2 — build + test (push) Has been cancelled
PR Test (sgl-router) / tier-3 — docker (placeholder) (push) Has been cancelled
PR Test (sgl-router) / tier-3 — k8s integration (push) Has been cancelled
PR Test (sgl-router) / tier-3 — e2e (push) Has been cancelled
PR Test (sgl-router) / finish (push) Has been cancelled
PR Test (NPU) / single-node-poc (map[name:qwen3_6_27b_w8a8_1p_in64k_out1k_50ms runner:linux-aarch64-a3-2 test_case:test/registered/ascend/performance/qwen3_6_27b/test_npu_qwen3_6_27b_w8a8_1p_in64k_out1k_50ms.py test_type:perf]) (push) Has been cancelled
PR Test (NPU) / pr-test-npu-finish (push) Has been cancelled
PR Test (Xeon) / pr-gate (push) Has been cancelled
PR Test (Xeon) / check-changes (push) Has been cancelled
PR Test (Xeon) / build-test (, xeon-gnr, base-b-test-cpu) (push) Has been cancelled
PR Test (XPU) / check-changes (push) Has been cancelled
PR Test (XPU) / pr-gate (push) Has been cancelled
PR Test (XPU) / stage-a-test-1-gpu-xpu (push) Has been cancelled
PR Test (XPU) / wait-for-stage-a (push) Has been cancelled
PR Test (XPU) / stage-b-test-1-gpu-xpu (push) Has been cancelled
PR Test (XPU) / finish (push) Has been cancelled
CI Model Inventory / build-inventory (push) Has been cancelled
Lint / lint (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark Compilation Check (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark - Manual Policy (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark - Request Processing (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark Summary (push) Has been cancelled
PR Test (SMG) / build-wheel (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on windows (x86_64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on macos (x86_64 - auto) (push) Has been cancelled
PR Test (SMG) / python-unit-tests (push) Has been cancelled
PR Test (SMG) / unit-tests (push) Has been cancelled
PR Test (SMG) / benchmarks (push) Has been cancelled
PR Test (SMG) / chat-completions (push) Has been cancelled
PR Test (SMG) / chat-completions-4gpu (push) Has been cancelled
PR Test (SMG) / e2e (push) Has been cancelled
PR Test (SMG) / docker-build-test (push) Has been cancelled
PR Test (SMG) / k8s-integration (push) Has been cancelled
PR Test (SMG) / finish (push) Has been cancelled
PR Test (SMG) / summarize-benchmarks (push) Has been cancelled
Release SGLang Model Gateway Docker Image / publish (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on macos (aarch64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (aarch64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (x86_64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (aarch64 - musllinux_1_1) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (x86_64 - musllinux_1_1) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / Build SDist (push) Has been cancelled
Release SGLang Model Gateway to PyPI / Upload to PyPI (push) Has been cancelled
Release SGLang Kernels / build-cu129-matrix (aarch64, 12.9, 3.10, arm-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / build-cu129-matrix (x86_64, 12.9, 3.10, x64-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / release-cu129 (push) Has been cancelled
Release SGLang Kernels / build-cu130-matrix (aarch64, 13.0, 3.10, arm-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / build-cu130-matrix (x86_64, 13.0, 3.10, x64-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / release-cu130 (push) Has been cancelled
Release SGLang Kernels / build-rocm-matrix (3.10, 700) (push) Has been cancelled
Release SGLang Kernels / build-rocm-matrix (3.10, 720) (push) Has been cancelled
Release SGLang Kernels / release-rocm700 (push) Has been cancelled
Release SGLang Kernels / release-rocm720 (push) Has been cancelled
Release SGLang Kernels / build-musa43 (43, 3.10) (push) Has been cancelled
Release SGLang Kernels / release-musa43 (push) Has been cancelled
217 lines
6.8 KiB
Python
217 lines
6.8 KiB
Python
import torch
|
|
import triton
|
|
import triton.language as tl
|
|
|
|
from sglang.kernels.ops.gemm.kernel_utils import _resolve_token_positions
|
|
from sglang.srt.lora.utils import LoRABatchInfo
|
|
|
|
|
|
@triton.jit
|
|
def _qkv_lora_b_kernel(
|
|
# Pointers to matrices
|
|
x,
|
|
weights,
|
|
output,
|
|
# Parameters of size
|
|
K, # K = R
|
|
max_qkv_out_dim, # max(output_q_dim, output_kv_dim)
|
|
# Strides
|
|
x_stride_0,
|
|
x_stride_1,
|
|
w_stride_0,
|
|
w_stride_1,
|
|
w_stride_2,
|
|
output_stride_0,
|
|
output_stride_1,
|
|
# Information on sequence lengths and weight id
|
|
seg_lens,
|
|
seg_indptr,
|
|
weight_indices,
|
|
lora_ranks,
|
|
# Offsets of q/k/v slice on output dimension
|
|
n_offs,
|
|
sorted_token_ids,
|
|
# Meta parameters
|
|
SORTED_BY_ADAPTER: tl.constexpr,
|
|
BLOCK_S: tl.constexpr,
|
|
BLOCK_N: tl.constexpr,
|
|
BLOCK_K: tl.constexpr,
|
|
# For fused output scaling
|
|
scalings,
|
|
):
|
|
"""
|
|
This kernel packs 3 sgemms (q/k/v) into a single kernel. The multiplication
|
|
results are accumulated into the output tensor.
|
|
|
|
When a sequence's rank is 0, the kernel is essentially a no-op, following
|
|
the convention in pytorch where the product of two matrices of shape (m, 0)
|
|
and (0, n) is an all-zero matrix of shape (m, n).
|
|
|
|
Args:
|
|
x (Tensor): The input tensor, which is the result of the LoRA A projection.
|
|
Shape: (s, 3 * K), where s is the sum of all sequence lengths in the
|
|
batch and K is the maximum LoRA rank. The second dimension is partitioned
|
|
for Q, K, and V.
|
|
weights (Tensor): The LoRA B weights for all adapters.
|
|
Shape: (num_lora, N_Q + 2 * N_KV, K).
|
|
output (Tensor): The output tensor where the result is stored.
|
|
Shape: (s, N_Q + 2 * N_KV).
|
|
"""
|
|
|
|
# Current block computes sequence with batch_id,
|
|
# which starts from row seg_start of x with length seg_len.
|
|
# qkv_id decides which of q,k,v to compute (0: q, 1: k, 2: v)
|
|
batch_id = tl.program_id(axis=2)
|
|
w_index = tl.load(weight_indices + batch_id)
|
|
rank = tl.load(lora_ranks + w_index)
|
|
|
|
# If rank is 0, this kernel is a no-op.
|
|
if rank == 0:
|
|
return
|
|
|
|
qkv_id = tl.program_id(axis=1)
|
|
pid = tl.program_id(axis=0)
|
|
seg_len = tl.load(seg_lens + batch_id)
|
|
if seg_len == 0:
|
|
return
|
|
seg_start = tl.load(seg_indptr + batch_id)
|
|
n_start = tl.load(n_offs + qkv_id)
|
|
n_size = tl.load(n_offs + qkv_id + 1) - n_start
|
|
scaling = tl.load(scalings + w_index)
|
|
# Adjust K (rank) according to the specific LoRA adapter
|
|
K = tl.minimum(K, rank)
|
|
|
|
# The tile in output matrix will have (pid_s, pid_n) as id
|
|
num_pid_n = tl.cdiv(max_qkv_out_dim, BLOCK_N)
|
|
pid_s = pid // num_pid_n
|
|
pid_n = pid % num_pid_n
|
|
if pid_s * BLOCK_S >= seg_len:
|
|
return
|
|
|
|
# Create pointers for the first block of x and weights[batch_id][n_start: n_end][:]
|
|
# The pointers will be advanced as we move in the K direction
|
|
# and accumulate
|
|
s_offset = tl.arange(0, BLOCK_S) + pid_s * BLOCK_S
|
|
n_offset = tl.arange(0, BLOCK_N) + pid_n * BLOCK_N
|
|
k_offset = tl.arange(0, BLOCK_K)
|
|
|
|
s_physical = _resolve_token_positions(
|
|
sorted_token_ids, seg_start, s_offset, seg_len, SORTED_BY_ADAPTER
|
|
)
|
|
x_ptrs = (
|
|
x
|
|
+ (qkv_id * K) * x_stride_1
|
|
+ (s_physical[:, None] * x_stride_0 + k_offset[None, :] * x_stride_1)
|
|
)
|
|
w_ptrs = (weights + w_index * w_stride_0 + n_start * w_stride_1) + (
|
|
k_offset[:, None] * w_stride_2 + n_offset[None, :] * w_stride_1
|
|
)
|
|
|
|
# Iterate to compute the block in output matrix
|
|
partial_sum = tl.zeros((BLOCK_S, BLOCK_N), dtype=tl.float32)
|
|
for k in range(0, tl.cdiv(K, BLOCK_K)):
|
|
x_tile = tl.load(
|
|
x_ptrs,
|
|
mask=(s_offset[:, None] < seg_len) & (k_offset[None, :] < K - k * BLOCK_K),
|
|
other=0.0,
|
|
)
|
|
w_tile = tl.load(
|
|
w_ptrs,
|
|
mask=(k_offset[:, None] < K - k * BLOCK_K) & (n_offset[None, :] < n_size),
|
|
other=0.0,
|
|
)
|
|
partial_sum += tl.dot(x_tile, w_tile)
|
|
|
|
x_ptrs += BLOCK_K * x_stride_1
|
|
w_ptrs += BLOCK_K * w_stride_2
|
|
|
|
# Store result to output matrix
|
|
partial_sum *= scaling
|
|
partial_sum = partial_sum.to(x.dtype.element_ty)
|
|
output_ptr = (
|
|
output
|
|
+ n_start * output_stride_1
|
|
+ (s_physical[:, None] * output_stride_0 + n_offset[None, :] * output_stride_1)
|
|
)
|
|
output_mask = (s_offset[:, None] < seg_len) & (n_offset[None, :] < n_size)
|
|
partial_sum += tl.load(output_ptr, mask=output_mask)
|
|
tl.store(output_ptr, partial_sum, mask=output_mask)
|
|
|
|
|
|
def qkv_lora_b_fwd(
|
|
x: torch.Tensor,
|
|
qkv_lora_b: torch.Tensor,
|
|
batch_info: LoRABatchInfo,
|
|
output_offset: torch.Tensor,
|
|
max_qkv_out_dim: int,
|
|
base_output: torch.Tensor = None,
|
|
n_slices: int = 3,
|
|
) -> torch.Tensor:
|
|
|
|
# x: (s, n_slices * r)
|
|
# qkv_lora_b: (num_lora, output_dim_q + 2 * output_dim_kv, r)
|
|
# output_offset = [0, output_dim_q, output_dim_q + output_dim_kv,
|
|
# output_dim_q + 2 * output_dim_kv] (length n_slices + 1)
|
|
# max_qkv_out_dim = max(output_dim_q, output_dim_kv)
|
|
# output: (s, output_dim_q + 2 * output_dim_kv)
|
|
|
|
# Compute lora_output with shape (s, output_dim) as follows:
|
|
# lora_output[:, :output_dim_q] = sgemm(x[:, :r], qkv_lora_b[:, :outptu_dim_q, :])
|
|
# lora_output[:, output_dim_q: output_dim_q + output_dim_kv]
|
|
# = sgemm(x[:, r: 2 * r], qkv_lora_b[:, outptu_dim_q: output_dim_q + output_dim_kv, :])
|
|
# lora_output[:, output_dim_q + output_dim_kv: ]
|
|
# = sgemm(x[:, 2 * r: , qkv_lora_b[:, output_dim_q + output_dim_kv: , :])
|
|
|
|
# Get dims
|
|
s = x.shape[0]
|
|
input_dim = x.shape[1]
|
|
r = qkv_lora_b.shape[-1]
|
|
output_dim = qkv_lora_b.shape[-2]
|
|
assert input_dim == n_slices * r
|
|
assert output_offset.shape[0] == n_slices + 1
|
|
|
|
BLOCK_S = 16
|
|
BLOCK_R = 16
|
|
BLOCK_OUT = 64
|
|
|
|
grid_b = (
|
|
triton.cdiv(batch_info.max_len, BLOCK_S)
|
|
* triton.cdiv(max_qkv_out_dim, BLOCK_OUT),
|
|
n_slices,
|
|
batch_info.bs,
|
|
)
|
|
|
|
if base_output is None:
|
|
output = torch.zeros((s, output_dim), device=x.device, dtype=x.dtype)
|
|
else:
|
|
output = base_output
|
|
|
|
sorted_by_adapter = batch_info.permutation is not None
|
|
_qkv_lora_b_kernel[grid_b](
|
|
x,
|
|
qkv_lora_b,
|
|
output,
|
|
r,
|
|
max_qkv_out_dim,
|
|
x.stride(0),
|
|
x.stride(1),
|
|
qkv_lora_b.stride(0),
|
|
qkv_lora_b.stride(1),
|
|
qkv_lora_b.stride(2),
|
|
output.stride(0),
|
|
output.stride(1),
|
|
batch_info.seg_lens,
|
|
batch_info.seg_indptr,
|
|
batch_info.weight_indices,
|
|
batch_info.lora_ranks,
|
|
output_offset,
|
|
batch_info.permutation,
|
|
sorted_by_adapter,
|
|
BLOCK_S,
|
|
BLOCK_OUT,
|
|
BLOCK_R,
|
|
batch_info.scalings,
|
|
)
|
|
|
|
return output
|